@@ -37,12 +37,54 @@
 #include "newformat.h"
 #include "acerrors.h"
 #include "attributes.h"
+#include "doio.h"
 
 #define ERROR(e) do { err = (e); goto err; } while (0)
 
+static int make_and_push_ext(AC *ac, char *name, char *data, int critical)
+{
+  X509_EXTENSION *ext = X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid(name), data);
+
+  if (ext) {
+    X509_EXTENSION_set_critical(ext, critical);
+    sk_X509_EXTENSION_push(ac->acinfo->exts, ext);
+    return 0;
+  }
+
+  X509_EXTENSION_free(ext);
+  return AC_ERR_NO_EXTENSION;
+}
+
+static void make_uri(const char *vo, const char *uri, STACK_OF(GENERAL_NAME) *names)
+{
+  GENERAL_NAME   *g    = NULL;
+  ASN1_IA5STRING *tmpr = NULL;
+
+  if (vo || uri) {
+    char *buffer=snprintf_wrap("%s://%s", vo ? vo : "", uri ? uri : "");
+
+    g = GENERAL_NAME_new();
+    tmpr = ASN1_IA5STRING_new();
+
+    if (!tmpr || !g || !buffer) {
+      GENERAL_NAME_free(g);
+      ASN1_IA5STRING_free(tmpr);
+      free(buffer);
+      return;
+    }
+
+    ASN1_STRING_set(tmpr, buffer, strlen(buffer));
+    free(buffer);
+    g->type  = GEN_URI;
+    g->d.ia5 = tmpr;
+    sk_GENERAL_NAME_push(names, g);
+  }
+}
+
 int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *pkey, BIGNUM *s,
             char **fqan, char *t, char **attributes_strings, AC **ac,
-            const char *vo, const char *uri, int valid, int old)
+            const char *vo, const char *uri, int valid, int old, int startpast,
+            STACK_OF(X509_EXTENSION) *extensions)
 {
   AC *a;
   X509_NAME *name1, *name2, *subjdup, *issdup;
@@ -52,32 +94,29 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
   AC_ATTR *capabilities;
   AC_IETFATTR *capnames;
   AC_FULL_ATTRIBUTES *ac_full_attrs;
-  ASN1_OBJECT *cobj, *aobj;
+  ASN1_OBJECT *cobj;
   X509_ALGOR *alg1, *alg2;
   ASN1_GENERALIZEDTIME *time1, *time2;
-  X509_EXTENSION *norevavail, *targets, *auth, *certstack;
-  AC_ATT_HOLDER *ac_att_holder;
-  char *qual, *name, *value, *tmp, *tmp2;
+  AC_ATT_HOLDER *ac_att_holder = NULL;
   STACK_OF(X509) *stk = NULL;
-
+  
   ASN1_NULL *null;
   int i = 0;
   int err = AC_ERR_UNKNOWN;
+  int ret = 0;
 
   time_t curtime;
 
   a = NULL;
-  name1 = name2 = subjdup = issdup = NULL;
+  subjdup = issdup = NULL;
   dirn = dirn2 = NULL;
-  version = serial = holdserial = NULL;
+  serial = holdserial = NULL;
   time1 = time2 = NULL;
   uid = NULL;
   capabilities = NULL;
   capnames = NULL;
-  cobj = aobj = NULL;
-  certstack = auth = targets = norevavail = NULL;
+  cobj = NULL;
   ac_full_attrs = NULL;
-  qual = name = value = tmp = tmp2 = NULL;
 
   if (!issuerc || !holder || !s || !fqan || !ac || !pkey)
     return AC_ERR_PARAMETERS;
@@ -99,8 +138,8 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
 
 
   time(&curtime);
-  time1 = ASN1_GENERALIZEDTIME_set(NULL, curtime);
-  time2 = ASN1_GENERALIZEDTIME_set(NULL, curtime+valid);
+  time1 = ASN1_GENERALIZEDTIME_set(NULL, curtime - startpast);
+  time2 = ASN1_GENERALIZEDTIME_set(NULL, curtime + valid - startpast);
 
   subjdup             = X509_NAME_dup(name2);
   issdup              = X509_NAME_dup(name1);
@@ -111,7 +150,6 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
   version             = BN_to_ASN1_INTEGER((BIGNUM *)(BN_value_one()), NULL);
   capabilities        = AC_ATTR_new();
   cobj                = OBJ_txt2obj("idatcap",0);
-  aobj                = OBJ_txt2obj("attributes",0);
   capnames            = AC_IETFATTR_new();
   null                = ASN1_NULL_new();
   ac_full_attrs       = AC_FULL_ATTRIBUTES_new();
@@ -122,52 +160,34 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
       !null || !ac_full_attrs || !ac_att_holder)
     ERROR(AC_ERR_MEMORY);
 
-  // prepare AC_IETFATTR
-  while(fqan[i]) 
-  {
+  /* prepare AC_IETFATTR */
+  while(fqan[i]) {
     ASN1_OCTET_STRING *tmpc = ASN1_OCTET_STRING_new();
-    if (!tmpc) 
-    {
+    if (!tmpc) {
       ASN1_OCTET_STRING_free(tmpc);
       ERROR(AC_ERR_MEMORY);
     }
-    ASN1_OCTET_STRING_set(tmpc, fqan[i], strlen(fqan[i]));
+    ASN1_OCTET_STRING_set(tmpc, (unsigned char*)fqan[i], strlen(fqan[i]));
     sk_AC_IETFATTRVAL_push(capnames->values, (AC_IETFATTRVAL *)tmpc);
     i++;
   }
-  {
-    GENERAL_NAME *g = GENERAL_NAME_new();
-    ASN1_IA5STRING *tmpr = ASN1_IA5STRING_new();
-    char *buffer=(char *)malloc(strlen(vo)+strlen(uri)+4);
 
-    if (!tmpr || !g || !buffer) 
-    {
-      GENERAL_NAME_free(g);
-      ASN1_IA5STRING_free(tmpr);
-      free(buffer);
-      ERROR(AC_ERR_MEMORY);
-    }
-    strcpy(buffer, vo);
-    strcat(buffer, "://");
-    strcat(buffer,uri);
-    ASN1_STRING_set(tmpr, buffer, strlen(buffer));
-    free(buffer);
-    g->type  = GEN_URI;
-    g->d.ia5 = tmpr;
-    sk_GENERAL_NAME_push(capnames->names, g);
+  if (vo || uri) {
+    make_uri(vo, uri, capnames->names);
+
+    /* stuff the created AC_IETFATTR in ietfattr (values) and define its object */
+    sk_AC_IETFATTR_push(capabilities->ietfattr, capnames);
+    capnames = NULL;
   }
 
-  // stuff the created AC_IETFATTR in ietfattr (values) and define its object
-  sk_AC_IETFATTR_push(capabilities->ietfattr, capnames);
   capabilities->get_type = GET_TYPE_FQAN;
   ASN1_OBJECT_free(capabilities->type);
   capabilities->type = cobj;
 
-
   i = 0;
-  // prepare AC_FULL_ATTRIBUTES
+  /* prepare AC_FULL_ATTRIBUTES */
   if (attributes_strings) {
-    while(attributes_strings[i]) {
+    while (attributes_strings[i]) {
       char *qual, *name, *value;
       char *tmp = NULL, *tmp2 = NULL;
 
@@ -178,7 +198,7 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
         ERROR(AC_ERR_MEMORY);
       }
 
-      tmp =strstr(attributes_strings[i], "::");
+      tmp = strstr(attributes_strings[i], "::");
       if (tmp == attributes_strings[i]) {
         qual = NULL;
         tmp = attributes_strings[i] + 2;
@@ -200,59 +220,42 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
       }
 
       if (qual)
-        ASN1_OCTET_STRING_set(ac_attr->qualifier, qual, strlen(qual));
+        ASN1_OCTET_STRING_set(ac_attr->qualifier, (unsigned char *)qual, strlen(qual));
+      else if (vo)
+        ASN1_OCTET_STRING_set(ac_attr->qualifier, (unsigned char *)vo, strlen(vo));
       else
-        ASN1_OCTET_STRING_set(ac_attr->qualifier, vo, strlen(vo));
+        ASN1_OCTET_STRING_set(ac_attr->qualifier, (unsigned char *)"", 0);
       
-      ASN1_OCTET_STRING_set(ac_attr->name,        name,  strlen(name));
-      ASN1_OCTET_STRING_set(ac_attr->value,       value, strlen(value));
+      ASN1_OCTET_STRING_set(ac_attr->name,        (unsigned char *)name,  strlen(name));
+      ASN1_OCTET_STRING_set(ac_attr->value,       (unsigned char *)value, strlen(value));
       
       sk_AC_ATTRIBUTE_push(ac_att_holder->attributes, ac_attr);
       i++;
     }
   }
+
   if (!i) 
     AC_ATT_HOLDER_free(ac_att_holder);
   else {
-    GENERAL_NAME *g = GENERAL_NAME_new();
-    ASN1_IA5STRING *tmpr = ASN1_IA5STRING_new();
-    char *buffer=(char *)malloc(strlen(vo)+strlen(uri)+4);
-
-    if (!tmpr || !g || !buffer) 
-    {
-      GENERAL_NAME_free(g);
-      ASN1_IA5STRING_free(tmpr);
-      free(buffer);
-      ERROR(AC_ERR_MEMORY);
-    }
-    strcpy(buffer, vo);
-    strcat(buffer, "://");
-    strcat(buffer,uri);
-    ASN1_STRING_set(tmpr, buffer, strlen(buffer));
-    free(buffer);
-    g->type  = GEN_URI;
-    g->d.ia5 = tmpr;
-    sk_GENERAL_NAME_push(ac_att_holder->grantor, g);
-
+    make_uri(vo, uri,  ac_att_holder->grantor);
     sk_AC_ATT_HOLDER_push(ac_full_attrs->providers, ac_att_holder);
   }  
   
-  // push both AC_ATTR into STACK_OF(AC_ATTR)
+  /* push both AC_ATTR into STACK_OF(AC_ATTR) */
   sk_AC_ATTR_push(a->acinfo->attrib, capabilities);
+  capabilities = NULL;
 
-  if (ac_full_attrs) {
-    X509_EXTENSION *ext = NULL;
-    ext=X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid("attributes"), (char *)(ac_full_attrs->providers));
+  if (ac_full_attrs && i) {
+    ret = make_and_push_ext(a, "attributes", (char *)(ac_full_attrs->providers), 0);
     AC_FULL_ATTRIBUTES_free(ac_full_attrs);
-
-    if (!ext)
-      ERROR(AC_ERR_NO_EXTENSION);
-
-    //    X509_EXTENSION_set_critical(targets,1);
-    sk_X509_EXTENSION_push(a->acinfo->exts, ext);
-    //    AC_FULL_ATTRIBUTES_free(ac_full_attrs);
     ac_full_attrs = NULL;
+    ac_att_holder = NULL;
+
+    if (ret)
+      ERROR(AC_ERR_NO_EXTENSION);
   }
+  else
+    AC_FULL_ATTRIBUTES_free(ac_full_attrs);
 
   stk = sk_X509_new_null();
 
@@ -269,34 +272,39 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
                (X509 *)ASN1_dup((int (*)())i2d_X509,(char * (*)())d2i_X509, (char *)issuerc));
 #endif
 
-  certstack = X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid("certseq"), (char*)stk);
+  ret = make_and_push_ext(a, "certseq", (char*)stk, 0);
   sk_X509_pop_free(stk, X509_free);
 
-  /* Create extensions */
-  norevavail=X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid("idcenoRevAvail"), "loc");
-  if (!norevavail)
+  if (ret)
     ERROR(AC_ERR_NO_EXTENSION);
-  X509_EXTENSION_set_critical(norevavail, 0); 
 
-  auth = X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid("authKeyId"), (char *)issuerc);
-  if (!auth)
+  /* Create several extensions */
+  if (make_and_push_ext(a, "idcenoRevAvail", "loc", 0))
     ERROR(AC_ERR_NO_EXTENSION);
-  X509_EXTENSION_set_critical(auth, 0); 
 
-  if (t) {
-    targets=X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid("idceTargets"), t);
-    if (!targets)
-      ERROR(AC_ERR_NO_EXTENSION);
+  if (make_and_push_ext(a, "authKeyId", (char *)issuerc, 0))
+    ERROR(AC_ERR_NO_EXTENSION);
 
-    X509_EXTENSION_set_critical(targets,1);
-    sk_X509_EXTENSION_push(a->acinfo->exts, targets);
+  if (t && make_and_push_ext(a, "idceTargets", t, 1))
+    ERROR(AC_ERR_NO_EXTENSION);
+
+  if (extensions) {
+    int proxyindex = 0;
+
+    for (proxyindex = 0; proxyindex < sk_X509_EXTENSION_num(extensions); proxyindex++) {
+      X509_EXTENSION *ext = X509_EXTENSION_dup(sk_X509_EXTENSION_value(extensions, i));
+      if (ext) {
+        if (!sk_X509_EXTENSION_push(a->acinfo->exts, ext)) {
+          X509_EXTENSION_free(ext);
+          goto err;
+        }
+      }
+      else {
+        goto err;
+      }
+    }
   }
 
-  sk_X509_EXTENSION_push(a->acinfo->exts, norevavail);
-  sk_X509_EXTENSION_push(a->acinfo->exts, auth);
-  if (certstack)
-    sk_X509_EXTENSION_push(a->acinfo->exts, certstack);
-
   alg1 = X509_ALGOR_dup(issuerc->cert_info->signature);
   alg2 = X509_ALGOR_dup(issuerc->sig_alg);
 
@@ -304,39 +312,35 @@ int writeac(X509 *issuerc, STACK_OF(X509) *issuerstack, X509 *holder, EVP_PKEY *
     if (!(uid = M_ASN1_BIT_STRING_dup(issuerc->cert_info->issuerUID)))
       ERROR(AC_ERR_MEMORY);
 
-  ASN1_INTEGER_free(a->acinfo->holder->baseid->serial);
-  ASN1_INTEGER_free(a->acinfo->serial);
-  ASN1_INTEGER_free(a->acinfo->version);
-  ASN1_GENERALIZEDTIME_free(a->acinfo->validity->notBefore);
-  ASN1_GENERALIZEDTIME_free(a->acinfo->validity->notAfter);
+#define FREE_AND_SET(datum, value, type) type##_free((datum)); (datum) = (value)
+
+  FREE_AND_SET(a->acinfo->holder->baseid->serial, holdserial, ASN1_INTEGER);
+  FREE_AND_SET(a->acinfo->serial, serial, ASN1_INTEGER);
+  FREE_AND_SET(a->acinfo->version, version, ASN1_INTEGER);
+  FREE_AND_SET(a->acinfo->validity->notBefore, time1, ASN1_GENERALIZEDTIME);
+  FREE_AND_SET(a->acinfo->validity->notAfter, time2, ASN1_GENERALIZEDTIME);
+  FREE_AND_SET(a->acinfo->alg, alg1, X509_ALGOR);
+  FREE_AND_SET(a->sig_alg, alg2, X509_ALGOR);
+
+#undef FREE_AND_SET
+
   dirn->d.dirn = subjdup;
   dirn->type = GEN_DIRNAME;
   sk_GENERAL_NAME_push(a->acinfo->holder->baseid->issuer, dirn);
   dirn2->d.dirn = issdup;
   dirn2->type = GEN_DIRNAME;
   sk_GENERAL_NAME_push(a->acinfo->form->names, dirn2);
-  a->acinfo->holder->baseid->serial = holdserial;
-  a->acinfo->serial = serial;
-  a->acinfo->version = version;
-  a->acinfo->validity->notBefore = time1;
-  a->acinfo->validity->notAfter  = time2;
   a->acinfo->id = uid;
-  X509_ALGOR_free(a->acinfo->alg);
-  a->acinfo->alg = alg1;
-  X509_ALGOR_free(a->sig_alg);
-  a->sig_alg = alg2;
 
   ASN1_sign((int (*)())i2d_AC_INFO, a->acinfo->alg, a->sig_alg, a->signature,
 	    (char *)a->acinfo, pkey, EVP_sha1());
 
   *ac = a;
   return 0;
- err:
 
-  X509_EXTENSION_free(auth);
-  X509_EXTENSION_free(norevavail);
-  X509_EXTENSION_free(targets);
-  X509_EXTENSION_free(certstack);
+ err:
+  sk_X509_EXTENSION_pop_free(a->acinfo->exts, X509_EXTENSION_free);
+  a->acinfo->exts = NULL;
   X509_NAME_free(subjdup);
   X509_NAME_free(issdup);
   GENERAL_NAME_free(dirn);
