@@ -44,6 +44,7 @@ extern "C" {
 #include <openssl/pem.h>
 #include <openssl/evp.h>
 #include "credentials.h"
+#include "sslutils.h"
 }
 
 #include <string>
@@ -64,84 +65,109 @@ extern "C" {
 #include "realdata.h"
 
 #include "internal.h"
+#include "normalize.h"
 
-static bool dncompare(const std::string &mut, const std::string &fixed);
+#ifndef VOMS_MAYBECONST
+#if defined(D2I_OF)
+#define VOMS_MAYBECONST const
+#else
+#define VOMS_MAYBECONST
+#endif
+#endif
+
+extern proxy_verify_desc *setup_initializers(char *cadir);
+extern void destroy_initializers(void *data);
+static bool dncompare(const char *mut, const char *fixed);
 static bool readdn(std::ifstream &file, char *buffer, int buflen);
 
 extern std::map<vomsdata*, vomsspace::internal*> privatedata;
 extern pthread_mutex_t privatelock;
 
-static void normalize(std::string &name)
+static bool dncompare(const char *first, const char *second)
 {
-  std::string::size_type pos = name.find(std::string("/USERID="));
-  while (pos != std::string::npos) {
-    name = name.substr(0, pos) + "/UID=" + name.substr(pos+8);
-    pos = name.find("/USERID=", pos+1);
-  }
-
-  pos = name.find(std::string("/emailAddress="));
-  while (pos != std::string::npos) {
-    name = name.substr(0, pos) + "/Email=" + name.substr(pos+14);
-    pos = name.find("/emailAddress=", pos+1);
-  }
-
-  pos = name.find(std::string("/E="));
-  while (pos != std::string::npos) {
-    name = name.substr(0, pos) + "/Email=" + name.substr(pos+3);
-    pos = name.find("/E=", pos+1);
-  }
-}
-
-static bool dncompare(const std::string &first, const std::string &second)
-{
-  if (first == second)
+  if (!strcmp(first, second))
     return true;
 
-  std::string s1 = first;
-  std::string s2 = second;
+  char *s1 = normalize(first);
+  char *s2 = normalize(second);
 
-  normalize(s1);
-  normalize(s2);
+  int res = strcmp(s1, s2);
 
-  return s1 == s2;
+  free(s1);
+  free(s2);
+
+  return res == 0;
 }
 
-
 bool
 vomsdata::evaluate(AC_SEQ *acs, const std::string& subject, 
                    const std::string& ca, X509 *holder)
 {
   bool ok = false;
 
-  if (holder) {
-    error = VERR_FORMAT;
+  error = VERR_FORMAT;
 
-    if (acs) {
-      /* Only new types. bn may or may not be set. */
-      int acnum = sk_AC_num(acs->acs);
+  if (acs) {
+    /* Only new types. bn may or may not be set. */
+    int acnum = sk_AC_num(acs->acs);
 
-      for (int i = 0; i < acnum; i++) {
-        ok = false;
-        voms v;
+    for (int i = 0; i < acnum; i++) {
+      ok = false;
+      voms v;
           
-        AC *ac = (AC *)sk_AC_value(acs->acs, i);
-        if (verifydata(ac, subject, ca, holder, v)) {
-          data.push_back(v);
-          ok = true;
-        }
-
-        if (!ok)
-          break;
+      AC *ac = (AC *)sk_AC_value(acs->acs, i);
+      if (verifydata(ac, subject, ca, holder, v)) {
+        data.push_back(v);
+        ok = true;
       }
+
+      if (!ok)
+        break;
     }
-    else
-      seterror(VERR_FORMAT, "AC not present in credentials.");
   }
+  else
+    seterror(VERR_FORMAT, "AC not present in credentials.");
 
   return ok;
 }
 
 
+static X509_EXTENSION *get_ext(X509 *cert, const char *name)
+{
+  int nid   = OBJ_txt2nid(name);
+  int index = X509_get_ext_by_NID(cert, nid, -1);
+
+  if (index >= 0)
+    return X509_get_ext(cert, index);
+  else
+    return NULL;
+}
+
+static bool findexts(X509 *cert , AC_SEQ **listnew, std::string &extra_data, std::string &workvo)
+{
+  X509_EXTENSION *ext;
+  bool found = false;
+
+  ext = get_ext(cert, "acseq");
+  if (ext) {
+    *listnew = (AC_SEQ *)X509V3_EXT_d2i(ext);
+    found = true;
+  }
+
+  ext = get_ext(cert, "incfile");
+  if (ext) {
+    extra_data = std::string((char *)(ext->value->data),ext->value->length);
+    found = true;
+  }
+
+  ext = get_ext(cert, "vo");
+  if (ext) {
+    workvo = std::string((char *)(ext->value->data),ext->value->length);
+  }
+
+  return found;
+}
+
 bool 
 vomsdata::retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how,
                    AC_SEQ **listnew, std::string &subject, std::string &ca, X509 **holder)
@@ -156,14 +182,8 @@ vomsdata::retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how,
   /*
    * check credential and get the globus name
    */
-  int chain_length;
-  X509_EXTENSION *ext;
-  int index = 0;
-  int nidf = 0, nidv = 0, nida = 0;
-  int position = 0;
-  char buf[1000];
-
-  ca = subject = "";
+  ca.clear();
+  subject.clear();
 
   X509 *h = get_real_cert(cert, chain);
   if (!h) {
@@ -178,8 +198,15 @@ vomsdata::retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how,
     return false;
   }
 
-  ca = std::string(X509_NAME_oneline(X509_get_issuer_name(*holder), buf, 1000));
-  subject = std::string(X509_NAME_oneline(X509_get_subject_name(*holder), buf, 1000));
+  char *buf = NULL;
+  buf = X509_NAME_oneline(X509_get_issuer_name(*holder), NULL, 0);
+  ca = std::string(buf ? buf : "" );
+  OPENSSL_free(buf);
+
+  buf = X509_NAME_oneline(X509_get_subject_name(*holder), NULL, 0);
+  subject = std::string(buf ? buf : "");
+  OPENSSL_free(buf);
+
   if (ca.empty() || subject.empty()) {
     seterror(VERR_IDCHECK, "Cannot discover CA name or DN from user's certificate.");
     return false;
@@ -187,39 +214,14 @@ vomsdata::retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how,
 
   /* object's nid */
 
-  nidf = OBJ_txt2nid("incfile");
-  nidv = OBJ_txt2nid("vo");
-  nida = OBJ_txt2nid("acseq");
-
-  /* seek for extensions in chain */
-
-  index = X509_get_ext_by_NID(cert, nida, -1);
-  if (index >= 0) {
-    ext = X509_get_ext(cert,index);
-    if (ext){
-      *listnew = (AC_SEQ *)X509V3_EXT_d2i(ext);
-      found = true;
-    }
-  }
-
-  index = X509_get_ext_by_NID(cert, nidf, -1);
-  if (index >= 0) {
-    ext = X509_get_ext(cert,index);
-    if (ext){
-      extra_data = std::string((char *)(ext->value->data),ext->value->length);
-      found = true;
-    }
-  }
-
-  index = X509_get_ext_by_NID(cert, nidv, -1);
-  if (index >= 0) {
-    ext = X509_get_ext(cert,index);
-    if (ext) {
-      workvo = std::string((char *)(ext->value->data),ext->value->length);
-      found = true;
-    }
-  }
+  found = findexts(cert, listnew, extra_data, workvo);
 
+  /*
+   * RECURSE_DEEP means find *all* extensions, even if they are
+   * superceded by newer ones.
+   *
+   * Because of this, the search cannot stop here but must continue.
+   */
   if (found && how != RECURSE_DEEP)
     return true;
   
@@ -227,40 +229,20 @@ vomsdata::retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how,
    * May need to travel up the chain.
    */
   if (how != RECURSE_NONE) {
+    int chain_length = sk_X509_num(chain);
+    int position = 0;
     
-    chain_length = sk_X509_num(chain);
-    
-    while (position < chain_length) {
-      
+    while (position < chain_length) {  
       cert = sk_X509_value(chain,position);
-      
-      index = X509_get_ext_by_NID(cert, nida, -1);
-      if (index >= 0) {
-        ext = X509_get_ext(cert, index);
-        if (ext){
-          *listnew = (AC_SEQ *)X509V3_EXT_d2i(ext);
-          found = true;
-        }
-      }
-      
-      index = X509_get_ext_by_NID(cert, nidf, -1);
-      if (index >= 0) {
-        ext = X509_get_ext(cert,index);
-        if (ext){
-	  extra_data = std::string((char *)(ext->value->data),ext->value->length);
-          found = true;
-        }
-      }
-      
-      index = X509_get_ext_by_NID(cert, nidv, -1);
-      if (index >= 0) {
-        ext = X509_get_ext(cert,index);
-        if (ext) {
-          workvo = std::string((char *)(ext->value->data),ext->value->length);
-          found = true;
-        }
-      }
 
+      found |= findexts(cert, listnew, extra_data, workvo);      
+
+      /*
+       * RECURSE_DEEP means find *all* extensions, even if they are
+       * superceded by newer ones.
+       *
+       * Because of this, the search cannot stop here but must continue.
+       */
       if (found && how != RECURSE_DEEP)
         return true;
       
@@ -272,80 +254,69 @@ vomsdata::retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how,
   return found;
 }
 
+static bool verifyID(X509 *cert, const std::string &server, const std::string &serverca)
+{
+  bool result = true;
+
+  /* check server subject */
+  char *bufsub  = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
+  char *bufiss  = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
+  if (!bufsub || !bufiss || 
+      strcmp(bufsub, server.c_str()) ||
+      strcmp(bufiss, serverca.c_str()))
+    result = false;
+
+  OPENSSL_free(bufsub);
+  OPENSSL_free(bufiss);
+
+  return result;
+}
+
 bool 
-vomsdata::verifydata(std::string &message, std::string subject, std::string ca, 
-		    X509 *holder, voms &v)
+vomsdata::verifydata(std::string &message, UNUSED(std::string subject), 
+                     UNUSED(std::string ca), 
+                     X509 *holder, voms &v)
 {
   error = VERR_PARAM;
-  if (message.empty() || subject.empty() || ca.empty() || !holder)
-    return false;
 
-  bool result = false;
+  if (message.empty())
+    return false;
 
   error = VERR_FORMAT;
 
-  unsigned char *str  = (unsigned char *)(const_cast<char *>(message.data()));
-  unsigned char *orig = str;
+  VOMS_MAYBECONST unsigned char *str  = (VOMS_MAYBECONST unsigned char *)(message.data());
+  VOMS_MAYBECONST unsigned char *orig = str;
 
   AC   *tmp    = d2i_AC(NULL, &str, message.size());
-  X509 *issuer = NULL;
-
-  if (ver_type & VERIFY_SIGN) {
-    issuer = check((void *)tmp);
-
-    if (!issuer) {
-      AC_free(tmp);
-      seterror(VERR_SIGN, "Cannot verify AC signature!");
-      return false;
-    }
-  }
 
   if (tmp) {
     size_t off = str - orig;
     message = message.substr(off);
 
-    result = verifyac(holder, issuer, tmp, verificationtime, v);
-    if (!result) {
-      //      seterror(VERR_VERIFY, "Cannot verify AC");
-      AC_free(tmp);
-      X509_free(issuer);
-      return false;
-    }
-    else {
-      ((struct realdata *)v.realdata)->ac = tmp;
-      tmp = NULL;
-    }
-    
-    if (result && (ver_type & VERIFY_ID)) {
-      char buf[2048];
-      /* check server subject */
-      if (strcmp(X509_NAME_oneline(X509_get_subject_name(issuer), buf,2048),
-                 v.server.c_str()) ||
-          strcmp(X509_NAME_oneline(X509_get_issuer_name(issuer), buf,2048),
-                 v.serverca.c_str())) {
-        seterror(VERR_SERVER, "Mismatch between AC signer and AC issuer");
-        result = false;
-      }
-    }
+    bool result = verifydata(tmp, subject, ca, holder, v);
+
+    AC_free(tmp);
+
+    return result;
   }
 
-  X509_free(issuer);
-  AC_free(tmp);   
-  
-  if (result)
-    v.holder = X509_dup(holder);
-  return result;
+  return false;
 }
 
-
 bool 
-vomsdata::verifydata(AC *ac, const std::string& subject, const std::string& ca, 
+vomsdata::verifydata(AC *ac, UNUSED(const std::string& subject), 
+                     UNUSED(const std::string& ca), 
                      X509 *holder, voms &v)
 {
   error = VERR_PARAM;
-  if (!ac || subject.empty() || ca.empty() || !holder)
+  if (!ac)
     return false;
 
+  if (!holder && (ver_type & VERIFY_ID)) {
+    error = VERR_NOIDENT;
+    return false;
+  }
+
   bool result = false;
 
   error = VERR_FORMAT;
@@ -372,22 +343,16 @@ vomsdata::verifydata(AC *ac, const std::string& subject, const std::string& ca,
   }
   
   if (result && (ver_type & VERIFY_ID)) {
-    char buf[2048];
-    /* check server subject */
-    if (strcmp(X509_NAME_oneline(X509_get_subject_name(issuer), buf,2048),
-               v.server.c_str()) ||
-        strcmp(X509_NAME_oneline(X509_get_issuer_name(issuer), buf,2048),
-               v.serverca.c_str())) {
+    if (!verifyID(issuer, v.server, v.serverca)) {
       seterror(VERR_SERVER, "Mismatch between AC signer and AC issuer");
       result = false;
     }
   }
 
-
   X509_free(issuer);
   
   if (result)
-    v.holder = X509_dup(holder);
+    v.holder = holder ? X509_dup(holder) : NULL;
   return result;
 }
 
@@ -412,14 +377,6 @@ bool vomsdata::check_sig_ac(X509 *cert, void *data)
   return (res == 1);
 }
 
-X509 *
-vomsdata::check(check_sig f, void *data)
-{
-  // This should not be used anymore.  Only left here for binary compatibility.
-  return NULL;
-}
-
-
 X509 *
 vomsdata::check(void *data)
 {
@@ -428,28 +385,28 @@ vomsdata::check(void *data)
   /* extract vo name from AC */
   
   AC * ac = (AC *)data;
-  STACK_OF(AC_ATTR) * atts = ac->acinfo->attrib;
-  int nid = OBJ_txt2nid("idatcap");
-  int pos = X509at_get_attr_by_NID(atts, nid, -1);
+  const STACK_OF(AC_ATTR) * atts = ac->acinfo->attrib;
 
-  int nidc = OBJ_txt2nid("certseq");
-  STACK_OF(X509_EXTENSION) *exts = ac->acinfo->exts;
-  int posc = X509v3_get_ext_by_NID(exts, nidc, -1);
+  int nid = OBJ_txt2nid("idatcap");
+  int pos = X509at_get_attr_by_NID((const STACK_OF(X509_ATTRIBUTE)*)atts, nid, -1);
 
   if (!(pos >=0)) {
     seterror(VERR_DIR, "Unable to extract vo name from AC.");
     return NULL;
   }
+
   AC_ATTR * caps = sk_AC_ATTR_value(atts, pos);
   if(!caps) {
     seterror(VERR_DIR, "Unable to extract vo name from AC.");
     return NULL;
   }
+
   AC_IETFATTR * capattr = sk_AC_IETFATTR_value(caps->ietfattr, 0);
   if(!capattr) {
     seterror(VERR_DIR, "Unable to extract vo name from AC.");
     return NULL;
   }
+
   GENERAL_NAME * name = sk_GENERAL_NAME_value(capattr->names, 0);
   if(!name) {
     seterror(VERR_DIR, "Unable to extract vo name from AC.");
@@ -464,7 +421,7 @@ vomsdata::check(void *data)
     std::string::size_type cpos2 = voname.find(":", cpos+1);
 
     if (cpos2 != std::string::npos) 
-      hostname = voname.substr(cpos+3, (cpos2 - cpos - 3));
+      hostname = voname.substr(cpos + 3, (cpos2 - cpos - 3));
     else {
       seterror(VERR_DIR, "Unable to determine hostname from AC.");
       return NULL;
@@ -477,15 +434,17 @@ vomsdata::check(void *data)
     return NULL;
   }
 
-
   /* check if the DN/CA file is installed for a given VO. */
+
+  int nidc = OBJ_txt2nid("certseq");
+  int posc = X509v3_get_ext_by_NID(ac->acinfo->exts, nidc, -1);
+
   if (posc >= 0) {
     std::string filecerts = voms_cert_dir + "/" + voname + "/" + hostname + ".lsc";
     std::ifstream file(filecerts.c_str());
 
     if (file)
       return check_from_file(ac, file, voname, filecerts);
-
   }
 
   /* check if able to find the signing certificate 
@@ -498,9 +457,9 @@ X509 *vomsdata::check_from_certs(AC *ac, const std::string& voname)
 {
   bool found  = false;
 
-  DIR * dp = NULL;
-  BIO * in = NULL;
-  X509 * x = NULL;
+  DIR  * dp = NULL;
+  BIO  * in = NULL;
+  X509 * x  = NULL;
 
   for(int i = 0; (i < 2 && !found); ++i) {
     
@@ -520,10 +479,13 @@ X509 *vomsdata::check_from_certs(AC *ac, const std::string& voname)
       char * name = de->d_name;
       if (name) {
         in = BIO_new(BIO_s_file());
+
         if (in) {
           std::string temp = directory + "/" + name;
+
           if (BIO_read_filename(in, temp.c_str()) > 0) {
             x = PEM_read_bio_X509(in, NULL, 0, NULL);
+
             if (x) {
               if (check_sig_ac(x, ac)) {
                 found = true;
@@ -544,8 +506,7 @@ X509 *vomsdata::check_from_certs(AC *ac, const std::string& voname)
     dp = NULL;
   }
 
-  if (in) 
-    BIO_free(in);
+  BIO_free(in);
   if (dp)
     (void)closedir(dp);
 
@@ -576,7 +537,6 @@ static bool readdn(std::ifstream &file, char *buffer, int buflen)
       return false;
 
     len = strlen(buffer);
-    int drop = 0;
     int start = 0;
     while (buffer[start] && isspace(buffer[start]))
       start++;
@@ -665,19 +625,22 @@ X509 *vomsdata::check_from_file(AC *ac, std::ifstream &file, const std::string &
       char issuercandidate[1000];
 
       X509 *current = sk_X509_value(certstack, i);
-      if (!readdn(file, subjcandidate, 1000) ||
-          !readdn(file, issuercandidate, 1000)) {
+      if (!readdn(file, subjcandidate, 999) ||
+          !readdn(file, issuercandidate, 999)) {
         success = false;
         final = true;
         break;
       }
 
+      subjcandidate[999] = issuercandidate[999] = '\0';
+
       char *realsubj = X509_NAME_oneline(X509_get_subject_name(current), NULL, 0);
       char *realiss  = X509_NAME_oneline(X509_get_issuer_name(current), NULL, 0);
       if (!dncompare(realsubj, subjcandidate) ||
           !dncompare(realiss, issuercandidate)) {
         do {
           file.getline(subjcandidate, 999);
+          subjcandidate[999] = '\0';
         } while (file && strcmp(subjcandidate, "------ NEXT CHAIN ------"));
         success = false;
         break;
@@ -691,6 +654,7 @@ X509 *vomsdata::check_from_file(AC *ac, std::ifstream &file, const std::string &
   } while (!final);
 
   file.close();
+
   if (!success) {
     AC_CERTS_free(certs);
     seterror(VERR_SIGN, "Unable to match certificate chain against file: " + filename);
@@ -726,38 +690,19 @@ X509 *vomsdata::check_from_file(AC *ac, std::ifstream &file, const std::string &
 bool
 vomsdata::check_cert(X509 *cert)
 {
-  X509_STORE *ctx = NULL;
-  X509_STORE_CTX *csc = NULL;
-  X509_LOOKUP *lookup = NULL;
-  int i = 0;
+  STACK_OF(X509) *stack = sk_X509_new_null();
 
-  csc = X509_STORE_CTX_new();
-  ctx = X509_STORE_new();
-  error = VERR_MEM;
-  if (ctx && csc) {
-    X509_STORE_set_verify_cb_func(ctx,cb);
-#ifdef SIGPIPE
-    signal(SIGPIPE,SIG_IGN);
-#endif
-    CRYPTO_malloc_init();
-    if ((lookup = X509_STORE_add_lookup(ctx, X509_LOOKUP_file()))) {
-      X509_LOOKUP_load_file(lookup, NULL, X509_FILETYPE_DEFAULT);
-      if ((lookup=X509_STORE_add_lookup(ctx,X509_LOOKUP_hash_dir()))) {
-        X509_LOOKUP_add_dir(lookup, ca_cert_dir.c_str(), X509_FILETYPE_PEM);
-        ERR_clear_error();
-        error = VERR_VERIFY;
-        X509_STORE_CTX_init(csc,ctx,cert,NULL);
-        if (verificationtime) {
-	  X509_STORE_CTX_set_time(csc, 0, verificationtime);
-        }
-        i = X509_verify_cert(csc);
-      }
-    }
+  if (stack) {
+    sk_X509_push(stack, cert);
+
+    bool result = check_cert(stack);
+
+    sk_X509_free(stack);
+
+    return result;
   }
-  if (ctx) X509_STORE_free(ctx);
-  if (csc) X509_STORE_CTX_free(csc);
 
-  return (i != 0);
+  return false;
 }
 
 bool
@@ -772,69 +717,49 @@ vomsdata::check_cert(STACK_OF(X509) *stack)
   ctx = X509_STORE_new();
   error = VERR_MEM;
   if (ctx && csc) {
-    X509_STORE_set_verify_cb_func(ctx,cb);
+    proxy_verify_desc *pvd = setup_initializers(strdup((char*)ca_cert_dir.c_str()));
+
+    X509_STORE_set_verify_cb_func(ctx,proxy_verify_callback);
+    
 #ifdef SIGPIPE
-    signal(SIGPIPE,SIG_IGN);
+    void (*oldsignal)(int) = signal(SIGPIPE,SIG_IGN);
 #endif
     CRYPTO_malloc_init();
     if ((lookup = X509_STORE_add_lookup(ctx, X509_LOOKUP_file()))) {
       X509_LOOKUP_load_file(lookup, NULL, X509_FILETYPE_DEFAULT);
+
       if ((lookup=X509_STORE_add_lookup(ctx,X509_LOOKUP_hash_dir()))) {
         X509_LOOKUP_add_dir(lookup, ca_cert_dir.c_str(), X509_FILETYPE_PEM);
+
         for (int i = 1; i < sk_X509_num(stack); i++)
           X509_STORE_add_cert(ctx,sk_X509_value(stack, i));
+
         ERR_clear_error();
         error = VERR_VERIFY;
         X509_STORE_CTX_init(csc, ctx, sk_X509_value(stack, 0), NULL);
+        X509_STORE_CTX_set_ex_data(csc, PVD_STORE_EX_DATA_IDX, pvd);
         index = X509_verify_cert(csc);
       }
     }
+#ifdef SIGPIPE
+    signal(SIGPIPE, oldsignal);
+#endif
+    destroy_initializers(pvd);
   }
-  if (ctx) X509_STORE_free(ctx);
-  if (csc) X509_STORE_CTX_free(csc);
+  X509_STORE_free(ctx);
+
+  if (csc)
+    X509_STORE_CTX_free(csc);
 
   return (index != 0);
 }
 
-static int MS_CALLBACK 
-cb(int ok, X509_STORE_CTX *ctx)
-{
-  char buf[256];
-
-  if (!ok) {
-    X509_NAME_oneline(X509_get_subject_name(ctx->current_cert),buf,256);
-    if (ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) ok=1;
-    /* since we are just checking the certificates, it is
-     * ok if they are self signed. But we should still warn
-     * the user.
-     */
-    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
-    /* Continue after extension errors too */
-    if (ctx->error == X509_V_ERR_INVALID_CA) ok=1;
-    if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) ok=1;
-    if (ctx->error == X509_V_ERR_INVALID_PURPOSE) ok=1;
-    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
-  }
-  return(ok);
-}
-
 bool
-vomsdata::my_conn(const std::string &hostname, int port, const std::string &contact,
-	int version, const std::string &command, std::string &u, std::string &uc,
-                  std::string &buf)
+vomsdata::contact(const std::string &hostname, int port, UNUSED(const std::string &contact),
+	const std::string &command, std::string &buf, std::string &u, std::string &uc,
+                  int timeout)
 {
-  return my_conn(hostname, port, contact, version, command, u, uc, buf, -1);
-}
-
-bool
-vomsdata::my_conn(const std::string &hostname, int port, const std::string &contact,
-	int version, const std::string &command, std::string &u, std::string &uc,
-                  std::string &buf, int timeout)
-{
-  GSISocketClient sock(hostname, port, version);
-
-  sock.RedirectGSIOutput(stderr);
-  sock.ServerContact(contact);
+  GSISocketClient sock(hostname, port);
 
   char *cacert = NULL;
   char *certdir = NULL;
@@ -878,9 +803,7 @@ vomsdata::my_conn(const std::string &hostname, int port, const std::string &cont
   }
   
   u  = sock.own_subject;
-  uc = sock.own_ca;
-  //  sc = sock.peer_ca;
-  //s  = sock.peer_subject;
+  uc.clear();
 
   if (u.empty()) {
     seterror(VERR_NOIDENT, sock.GetError());
@@ -894,30 +817,19 @@ vomsdata::my_conn(const std::string &hostname, int port, const std::string &cont
     return false;
   }
 
-  if (!sock.Receive(buf)) {
-    seterror(VERR_COMM, sock.GetError());
-    sock.Close();
-    return false;
-  }
+  std::string msg;
+  bool ret;
+
+  do {
+    ret = sock.Receive(msg);
+    if (!ret) {
+      seterror(VERR_COMM, sock.GetError());
+      sock.Close();
+      return false;
+    }
+    buf += msg;
+  } while (ret && ! msg.empty());
 
   sock.Close();
   return true;
 }
-
-bool
-vomsdata::contact(const std::string &hostname, int port, const std::string &contact,
-	const std::string &command, std::string &buffer, std::string &username,
-                  std::string &ca, int timeout)
-{
-  return my_conn(hostname, port, contact, globus(0), command, username, ca,
-                 buffer, timeout);
-}   
-
-bool
-vomsdata::contact(const std::string &hostname, int port, const std::string &contact,
-	const std::string &command, std::string &buffer, std::string &username,
-                  std::string &ca)
-{
-  return my_conn(hostname, port, contact, globus(0), command, username, ca,
-                 buffer, -1);
-}   
